/*         PLS.PRG                                     RAINER SCHLITTGEN  8.2.2013
**
**
**  Programs for pls estimation and segmentation
**
**   Procedure                                                                                      Line
**   ____________________________________________________________________________________________________
**   plspfad       : estimation of pls path model                                                     24
**   plspfadresid  : determination of residuals of a pls path model                                  191
**   clusterpls    : clustering objects according to pls path models                                 275
**   clusterplsThread : same as clusterpls but with 4 threads                                        431
**   geneticpls    : computing a startpartition for clustering objects according to pls path models  641
**   popfit        : help function for geneticpls                                                    742 
**   geneticplsThread :  same as  geneticpls but with 4 thread                                       823 
**
*/

  declare matrix   _pn != 75;               // population size            
  declare matrix   _pC != 0;                // recombination probability 
  declare matrix   _pM != 0.3;              // mutation probability      
  declare matrix _limit != 1000;            // number of generations        
  declare matrix  _number != 2000;          // number of subsamples for clusterpls   

/*
**> plspfad 
**  format: {Bhat,eta,what,lambda,resa,resi,R2,ret} = plspfad(dat,B,refl)
**  input: dat = (n,p) Datenmatrix 
**         B   (G,G) lower triangular matrix describing the interrelations of the latent 
**             variables: 
**             b_ij=+1 if eta_i and eta_j are positively correlated
**             b_ij=-1 if eta_i and eta_j are negatively correlated
**             b_ij= 0 if eta_i does not depend on eta_j in a direct way
**             (b_ii = 0 !)
**         refl   (p,1) vector describing with which latent variable the manifest variables
**                      are related.
**                      signed integers from 1 to G; g at position i means, that the i-th manifest 
**                      variable is connected with the g-th latent variable  
**                      positive: formativly, negative: reflectively   
**  output: Bhat (G,G) lower triangular matrix with the estimated coefficients of the inner model 
**          what (p,1) vector of outer weights   
**          lambda (p,1) vector with the loadings 
**          eta  (n,G)-matrix, the scores of the latent variables   
**          resa (n,?) matrix of residuals from outer model 
**          resi (n,?) matrix of residuals from inner model 
**          R2 (G,1) vector of R^2 values of regressions  
**          ret scalar, return code: 
**              0 normal convergence
**              1 limit of iterations attained without convergence       
** remarks: - reflective relations lead to mode A, formative to mode B to determine the weights         
**          - loadings lambda for non-standardized data  
**      
*/

proc (8)= plspfad(dat,B,refl);

local c,mdat,d,d1,g,GG,ind,iind,ind1,ind2,eta,walt,w1,i,p,ii,n,lambda,vorz,w;
local b1,resa,resi,r1,pre,it,R2,ret,fr;

p = cols(dat); 
n = rows(dat);
GG = rows(B);                                    // GG = number of latent variables
ind=seqa(1,1,p);
ind1=seqa(1,1,GG);
eta = zeros(n,GG);
lambda = zeros(p,1);
resa={};
resi={};
R2 = miss(zeros(GG,1),0);
ret = 0;

if (rows(refl) /= p); print "length of refl does not match number of columns of data"; stop; endif;
 
                                                    
/* 0. step: centering and norming of the data */

  mdat = meanc(dat);                               // standardisation of the data   
  d = dat-mdat';                                    
  d = d./stdc(d)'; 
      
 

/* 1. step: choice of weights */

w = ones(p,1);                                   //  starting vector of outer weights
// for g(1,GG,1); 
//  i = selif(ind,abs(refl) .== g); 
//  w[maxc(i)] = -1;
// endfor;  

/* 2. step: outer approximation */

for g (1,GG,1);
  i = (g .== abs(refl)); 
  iind = selif(ind,i);                              // indices of manif. variables connected with  
  eta[.,g] = d[.,iind]*w[iind];                     // latent variable g 
  eta[.,g] = eta[.,g]/stdc(eta[.,g]);               // standardisation
  c = corrx(eta[.,g]~d[.,iind]);
  c = c[2:(cols(d[.,iind])+1),1];
  c = sumc((c.>0)-(c.<0));
  if c<0; eta[.,g] = -1*eta[.,g]; endif;            // mainly positiv correlations
endfor;  

/*   main loop; comparison of w and walt */ 

walt = 0;
it = 0;
do while ((sumc(abs(w-walt))>.0001)  and (it < 40));   // start of main loop 
 it = it+1;                                                

/* 3. step: saving of w at walt */ 

 walt = w;

/* 4. step: inner approximation */

 vorz = vcx(eta); 
 vorz = (vorz.>0)-(vorz.<0); 
 eta = eta*(vorz .* (B+B'));  
 eta = eta./stdc(eta)';

/* 5. step: recomputation of outer weights */
 
 for g (1,GG,1);
   i = (g.==abs(refl));
   iind = selif(ind,i);                           // indices of manif. variables connected with       
                                                  // latent variable g                            
   if refl[iind] > 0;                             // mode B for formative relations 
    w[iind] = olsqr(eta[.,g],d[.,iind]);   
   else;                                          // mode A otherwise
    for  ii (1,rows(iind),1); 
      w[iind[ii]] = olsqr(d[.,iind[ii]],eta[.,g]);    
    endfor; 
   endif; 
 endfor;  
 
/* 6. step: recomputation of the scores of the latent variables */

 for g (1,GG,1); 
   i = (g.==abs(refl));
   iind = selif(ind,i);                           // indices of manif. variables connected with        
                                                  // latent variable g                            
   eta[.,g] = d[.,iind]*w[iind];
   eta[.,g] = eta[.,g]/stdc(eta[.,g]); 
   c = corrx(eta[.,g]~d[.,iind]);
   c = c[2:(rows(iind)+1),1];
   c = sumc((c.>0)-(c.<0));
   if c<0; eta[.,g]=-1*eta[.,g]; endif;
 endfor;   
  
endo;                                            // end of main loop         

/* determination of loadings lambda */

  d = dat-mdat';                                 // loadings for non-standardized data
  
for g (1,GG,1); 
  iind = selif(ind,g .== abs(refl));             // indices of manif. variables connected with   
                                                 // latent variable g                            
  if refl[iind] > 0;                             // formativ groups
    { b1,r1,pre } = olsqr2(eta[.,g], d[.,iind]);   
    lambda[iind] = b1;  
       resa = resa~r1; 
  else;                                          // reflective groups  
   for ii (1,rows(iind),1);
    { b1,r1,pre } = olsqr2(d[.,iind[ii]],eta[.,g]);
    lambda[iind[ii]] = b1;  
     resa = resa~r1; 
   endfor;
  endif;
endfor;   

/* determination of path coefficients beta */

for g (1,GG,1); 
 if not(B[g,.] == 0); 
   ii = selif(ind1,B[g,.]'./=0); 
   { b1,r1,pre } =  olsqr2(eta[.,g],eta[.,ii]);
   R2[g] = vcx(pre)/vcx(eta[.,g]);
   B[g,ii] = b1';  
   resi = resi~r1; 
 endif;
endfor;     

if ((sumc(abs(w-walt))>.0001)  and (it >= 40)) ; ret = 1; endif;
 
retp(B,eta,w,lambda,resa,resi,R2,ret);
endp;  

/*
**> plspfadresid 
**  format: {resa,resi} = plspfadresid(dat,B,lambda,refl,w) 
**  input: dat = (n,p) Datenmatrix 
**         B   (G,G) lower triangular matrix with coefficient for the interrelations of the latent 
**             variables 
**         lambda  (p,1) vector of loadings of measurement model
**         refl   (p,1) vector describing with which latent variable the manifest variables
**                      are related.
**                      signed integers from 1 to G; g at position i means, that the i-th manifest 
**                      variable is connected with the g-th latent variable  
**                      positive: formativly negative: reflectively   
**          w    (p,1) vector of weights for computation of scores of latent variables
**  output: resa (n,?) matrix of residuals from outer model 
**          resi (n,?) matrix of residuals from inner model 
**  remarks: reflective relations lead to mode A, formative to mode B to determine the weights         
**      
**      
*/

proc (2) = plspfadresid(dat,B,lambda,refl,w);

local c,d,d1,eta,g,GG,ind,iind,ind1,ind2,md,mdat,walt,w1,i,p,ii,n,vorz;
local b1,resa,resi,r1,pre,it,ret,Desmat;

Desmat = (B.>0)-(B.<0);
p = cols(dat);
n = rows(dat);
GG = rows(B);                                     // GG = number of latent variables
ind=seqa(1,1,p);
ind1=seqa(1,1,GG);
eta = zeros(n,GG);
lambda = zeros(p,1);
resa={};
resi={};
ret = 0;
                                             
/* 0. step: centering of the data */


  mdat = meanc(dat);                               // centering of the data   
  d = dat-mdat';     
 
/* computation of the scores of the latent variables */

 for g (1,GG,1);  
   iind = selif(ind,g.==abs(refl));               // indices of manif. variables connected with        
                                                  // latent variable g                            
   eta[.,g] = d[.,iind]*w[iind];
   eta[.,g] = eta[.,g]/stdc(eta[.,g]); 
   c = corrx(eta[.,g]~d[.,iind]);
   c = c[2:(rows(iind)+1),1];
   c = sumc((c.>0)-(c.<0));
   if c<0; eta[.,g]=-1*eta[.,g]; endif;
 endfor;    

/* determination of residuals of outer model */

for g (1,GG,1); 
  iind = selif(ind,g.==abs(refl));                // indices of manif. variables connected with   
                                                  // latent variable g                            
  if refl[iind] > 0;                              // formative groups
    r1 = eta[.,g] - d[.,iind]*lambda[iind];   
    resa = resa~r1; 
  else;                                          // reflective groups   
    r1 = d[.,iind]-eta[.,g].*lambda[iind]'; 
    resa = resa~r1;  
  endif;
endfor;   

/* determination of residuals of inner model */

for g (1,GG,1); 
 if not(B[g,.]==0); 
   ii = selif(ind1,B[g,.]'./=0); 
   r1  = eta[.,g] - eta[.,ii]*B[g,ii]'; 
   resi = resi~r1; 
 endif;
endfor;     
 
retp(resa,resi);
endp; 

/*
**> clusterpls
**
**   Purpose:    clustering objects in that way that each cluster has a its own 
**               set of coefficients in the pls-path model
**
**   Format:     {member,B2,coeff,r2,it} = clusterpls(dat,B,refl,k,start) 
**   Input:      dat   (n,p)-matrix, the values of the manifest variables 
**               B   (G,G) lower triangular matrix describing the interrelations of the latent 
**                   variables: 
**                   b_ij=+1 if eta_i and eta_j are positively correlated
**                   b_ij=-1 if eta_i and eta_j are negatively correlated
**                   b_ij= 0 if eta_i does not depend on eta_j in a direct way
**                   (b_ii = 0 !) 
**               refl   (p,1) vector describing with which and in which way the latent variables
**                            and the manifest variables are related.
**                            signed integers from 1 to G; +-g at position i means, that the i-th 
**                            manifest variable is connected with the g-th latent variable  
**                            - : reflectively;  + : formatively 
**               k   scalar, the number of clusters to be found 
**               start (n,1) vector of indicator of membership to start with or
**                      0   to start with a random assignment           
**   Output:     member (n,1)-vector, indicator of membership
**               B1     (k,G,G)-array, the path coefficients of the clusters 
**               coeff  (p,k)-matrix, the loadings of the clusters 
**               r2     scalar, the sum of squared residuals of the models.
**               it     number of iterations for exchanging objects  
**   Remarks:    - only for full reflective or formative-reflective or only formative relations between manifest 
**               and latent variables 
**               - investigation in descending order of size of sum of squared residuals only up to n/2 of observations
*/


proc (5) = clusterpls(dat,B,refl,k,start); 

local a,c,coeff,dat1,i,ind,indmat,ind1,it,j,m,member,memzahl;
local n,p,pred,ralt,rneu,r2,ssalt,ssneu,verschieb,z; 
local g,b1,b2,eta,w,lambda,resa,resi,ret,reti,indj,ralteinzel,indneu,indneu1,rneueinzel,fr;
local minzahl,ret1,laenge,stich,number,gg,res1,coeffs,membera,w1,ind11;
local tabu,la,zentren,eta1,i1,res2,dist1,h,R2out;

n = rows(dat);
p = cols(dat);                                    // number of manifest variables    
ind = seqa(1,1,n); 
g = rows(B);                                      // number of latent variables
minzahl = minc(sumc(abs(refl).==seqa(1,1,k)')); 
coeffs = {};   
res2 = zeros(n,k);
tabu = zeros(4,1); 

member = start; 

 
/*  If there is no vector to start with: random assignment   */    
    
if start == 0;    
  member = rempiric(n,1,seqa(1,1,k),1);   
endif;

/* estimation of separate models for the k clusters */  

ralt = zeros(k,1);
rneu = zeros(k,1);
reti = zeros(k,1);
ret1 = {};
ralteinzel = zeros(n,1); 
j=0;                                                 
do while j<k;
  j = j+1;
  indj = selif(ind,member.==j); 
  dat1 = dat[indj,.]; 
  ret = 1;
  if rows(dat1) >  minzahl;
    {B1,eta,w,lambda,resa,resi,R2out,ret} = plspfad(dat1,B,refl);  
    ralteinzel[indj] = sumc((resi')^2);
    ralt[j] = sumc(vec(resi^2)); 
   endif;  
  ret1 = ret1|ret;
endo;  

ssalt = sumc(ralt);  
verschieb = 1;  
rneueinzel = ralteinzel;

/*   reassigning of the objects to reduce the sum of squared residuals */

it=0;  
do while ((it < 100) and (verschieb > 0));  
   it=it+1;  
   verschieb = 0;
   ralteinzel = rneueinzel;  
   rneueinzel = sumc(ralteinzel)*ones(n,1);
   indneu = rev(sortc(ind~rneueinzel,2));
   indneu = indneu[.,1]; 
   j=0;                                                   // passing through the data 
   do while (j < (n/2));                                  // only the bad half will be examined 
    j = j+1;
    z = dat[indneu[j],.];                                 // object j which may be reassigned
    memzahl = sumc(member.==seqa(1,1,k)');
    a = member[indneu[j]]; 
    if (memzahl > minzahl) and scalmiss(indnv(indneu[j],tabu));     // clusters must contain >=minzahl objects   
       indneu1 = selif(ind,(member.==a) .and (ind ./= indneu[j]));
       dat1 = dat[indneu1,.];        
       {B1,eta,w,lambda,resa,resi,R2out,ret} = plspfad(dat1,B,refl); 
       rneu[a] = sumc(vec(resi^2));
       reti[a] = ret;
       rneueinzel[indneu1] = sumc(resi'^2);
       i = 0;
       do while i<k;
         i= i+1;
         if (i ne a);
           indneu1 = selif(ind,(member.==i) + (ind .== indneu[j]));
           dat1 = dat[indneu1,.]; 
           {B1,eta,w,lambda,resa,resi,R2out,ret} = plspfad(dat1,B,refl); 
           rneu[i] = sumc(vec(resi^2)); 
           reti[i] = ret;
         if sumc(sumc(resi'^2)) < sumc(rneueinzel[indneu1]);
            rneueinzel[indneu1] = sumc(resi'^2);
           endif;  
         endif;
       endo;
          
       indmat = eye(k);
       indmat[a,.] = ones(1,k); 
       ssneu = sumc(indmat.*rneu + (1-indmat).*ralt);
       ssneu[a] = ssalt;
       m = minc(ssneu);  
       ind1 = minindc(ssneu);        
      
       if ((m < ssalt) and (reti == 0));                  // if SS becomes smaller and reassigning    
                                                          // does not cause problems
        member[indneu[j]] = ind1;
        verschieb = 1;
        tabu = tabu[2:4]|indneu[j]; 
        ssalt = m;  
       endif;
    endif;
   endo;  
endo; 

B2 =  arrayinit(k|g|g,miss(0,0));  
coeff = ones(rows(lambda),k)*miss(0,0);
r2 = 0;
for i(1,k,1); 
 dat1 = selif(dat,member .== i);  
 if rows(dat1) > p ; 
   {B1,eta,w,lambda,resa,resi,R2out,ret} = plspfad(dat1,B,refl); 
   coeff[.,i] = lambda;
   B2[i,.,.] = B1;
   r2 = r2+sumc(vec(resi^2));
 endif;
endfor;

retp(member,B2,coeff,r2,it);
endp;

/*
**> clusterplsThread : same as clusterpls but with 4 threads
** 
*/ 

proc (6) = clusterplsThread(dat,B,refl,k,start); 

local a,c,coeff,dat1,i,ind,indmat,ind1,it,j,m,member,memzahl;
local n,p,pred,ralt,rneu,r2,ssalt,ssneu,verschieb,z; 
local g,b1,b2,eta,w,lambda,resa,resi,ret,reti,indj,ralteinzel,indneu,indneu1,rneueinzel,fr;
local minzahl,ret1,laenge,stich,number,gg,res1,coeffs,membera,w1,ind11;
local tabu,la,zentren,eta1,i1,res2,dist1,h;
local indneu11,dat11,indneu12,dat12,indneu13,dat13,indneu14,dat14,rneu1,ret2,rneu2,ret3,rneu3,ret4,rneu4;
local B11,lambda1,resa1,resi1, 
      B12,eta2,w2,lambda2,resa2,resi2,   
      B13,eta3,w3,lambda3,resa3,resi3, 
      B14,eta4,w4,lambda4,resa4,resi4,thread1,thread2,thread3,thread4,jj;
local R2out,R2out1,R2out2,R2out3,R2out4,startzeit,endzeit;       

startzeit = (3600~60~1~0.01)*time;  

n = rows(dat);
p = cols(dat);                                    // number of manifest variables    
ind = seqa(1,1,n); 
g = rows(B);                                      // number of latent variables
minzahl = minc(sumc(abs(refl).==seqa(1,1,k)')); 
coeffs = {};   
res2 = zeros(n,k);
tabu = zeros(4,1);
 
member = start;


/*  
    Falls kein Startvektor vorgegeben: Zufllige Zuordnung       
*/    
    
if start == 0;  
  
  member = rempiric(n,1,seqa(1,1,k),1);  
 
endif;

/* estimation of separate models for the k clusters */  

ralt = zeros(k,1);
rneu = zeros(k,1);
reti = zeros(k,1);
ret1 = {};
ralteinzel = zeros(n,1); 
j=0;                                                 
do while j<k;
  j = j+1;
  indj = selif(ind,member.==j); 
  dat1 = dat[indj,.]; 
  ret = 1;
  if rows(dat1) >  minzahl;
    {B1,eta,w,lambda,resa,resi,R2out,ret} = plspfad(dat1,B,refl);   
    ralteinzel[indj] = sumc((resi')^2);
    ralt[j] = sumc(vec(resi^2)); 
   endif;  
  ret1 = ret1|ret;
endo;  

ssalt = sumc(ralt); 

verschieb = 1; 
 
rneueinzel = ralteinzel;

/*
   reassigning of the objects to reduce the sum of squared residuals
*/

it=0;  
do while ((it < 100) and (verschieb > 0));  
   it=it+1;  
   verschieb = 0;
   ralteinzel = rneueinzel;  
   rneueinzel = sumc(ralteinzel)*ones(n,1); 
   indneu = rev(sortc(ind~ralteinzel,2));
   indneu = indneu[.,1]; 
   j=0;                                                   // passing through the data 
   do while (j < (n/2));                                  // only the bad half will be examined
    j = j+1;
    z = dat[indneu[j],.];                                 // object j which may be reassigned
    memzahl = sumc(member.==seqa(1,1,k)');
    a = member[indneu[j]]; 
    if (memzahl > minzahl) and scalmiss(indnv(indneu[j],tabu));   // clusters must contain 
                                                                  // >= minzahl objects    
       ret = {};
       rneu = {};       
       i = -3;       
       do while i <= (k-4);  
          i = i+4;
          ret1 = {};        ret2 = {};        ret3 = {};        ret4 = {};        
          rneu1 = {};       rneu2 = {};       rneu3 = {};       rneu4 = {};  
          thread1 = 0;      thread2 = 0;      thread3 = 0;      thread4 = 0;
          ThreadBegin; 
          if (k >= i);
            thread1 = 1;  
            if (i == a);
             indneu11 = selif(ind,(member.==a) .and (ind ./= indneu[j]));
            else;
             indneu11 = selif(ind,(member.==i) + (ind .== indneu[j]));
            endif;
            dat11 = dat[indneu11,.]; 
            {B11,eta1,w1,lambda1,resa1,resi1,R2out1,ret1} = plspfad(dat11,B,refl);
            rneu1 = sumc(vec(resi1'^2));  
          endif;     
          ThreadEnd;  
          ThreadBegin; 
          if (k >= (i+1)); 
            thread2 = 1;
            if ((i+1) == a);
             indneu12 = selif(ind,(member.==a) .and (ind ./= indneu[j]));
            else;
             indneu12 = selif(ind,(member.==(i+1)) + (ind .== indneu[j])); 
            endif; 
            dat12 = dat[indneu12,.]; 
            {B12,eta2,w2,lambda2,resa2,resi2,R2out2,ret2} = plspfad(dat12,B,refl); 
            rneu2 = sumc(vec(resi2'^2)); 
          endif; 
          ThreadEnd; 
          ThreadBegin;  
          if (k >= (i+2));   
             thread3 = 1; 
             if ((i+2) == a);
              indneu13 = selif(ind,(member.==a) .and (ind ./= indneu[j]));
             else; 
              indneu13 = selif(ind,(member.==(i+2)) + (ind .== indneu[j]));
             endif;
             dat13 = dat[indneu13,.]; 
             {B13,eta3,w3,lambda3,resa3,resi3,R2out3,ret3} = plspfad(dat13,B,refl); 
             rneu3 = sumc(vec(resi3'^2)); 
          endif;  
          ThreadEnd; 
          ThreadBegin;  
          if (k >= (i+3));
             thread4 = 1; 
             if ((i+3) == a);
              indneu14 = selif(ind,(member.==a) .and (ind ./= indneu[j]));
             else;
              indneu14 = selif(ind,(member.==(i+3)) + (ind .== indneu[j]));
             endif; 
             dat14 = dat[indneu14,.]; 
             {B14,eta4,w4,lambda4,resa4,resi4,R2out4,ret4} = plspfad(dat14,B,refl); 
             rneu4 = sumc(vec(resi4'^2));  
          endif; 
          ThreadEnd; 
          ThreadJoin;
          rneu = rneu|rneu1|rneu2|rneu3|rneu4;
          ret = ret|ret1|ret2|ret3|ret4;
          if thread1 == 1;
            if (sumc(sumc(resi1'^2)) < sumc(rneueinzel[indneu11]));
               rneueinzel[indneu11] = sumc(resi1'^2);
            endif;  
          endif;  
          if thread2 == 1;
           if (sumc(sumc(resi2'^2)) < sumc(rneueinzel[indneu12]));
               rneueinzel[indneu12] = sumc(resi2'^2);
           endif; 
          endif; 
          if thread3 == 1;
           if  (sumc(sumc(resi3'^2)) < sumc(rneueinzel[indneu13]));
               rneueinzel[indneu13] = sumc(resi3'^2);
           endif; 
          endif; 
          if thread4 == 1;
           if  (sumc(sumc(resi4'^2)) < sumc(rneueinzel[indneu14]));
               rneueinzel[indneu14] = sumc(resi4'^2);
           endif;  
          endif;  
       endo; 
          
       indmat = eye(k);
       indmat[a,.] = ones(1,k); 
       ssneu = sumc(indmat.*rneu + (1-indmat).*ralt);
       ssneu[a] = ssalt;
       m = minc(ssneu);  
       ind1 = minindc(ssneu);     
       if ((m < ssalt) and (reti == 0));                  // if SS becomes smaller and reassigning    
                                                          // does not cause problems
        member[indneu[j]] = ind1;
        verschieb = 1;
        tabu = tabu[2:4]|indneu[j]; 
        ssalt = m;  
       endif;
    endif;
   endo;  
endo; 

B2 =  arrayinit(k|g|g,miss(0,0));  
coeff = ones(rows(lambda),k)*miss(0,0);
r2 = 0;
for i(1,k,1); 
 dat1 = selif(dat,member .== i);  
 if rows(dat1) > p ; 
   {B1,eta,w,lambda,resa,resi,R2out,ret} = plspfad(dat1,B,refl); 
   coeff[.,i] = lambda;
   B2[i,.,.] = B1;
   r2 = r2+sumc(vec(resi^2));
 endif;
endfor;

endzeit = (3600~60~1~0.01)*time; 

retp(member,B2,coeff,r2,it,endzeit-startzeit);
endp;

/*
**> geneticpls
**
**   Purpose:    computing a startpartition for clustering objects in that way that each 
**               cluster has a its own regression relation
** 
**   Format:     {member,fitness,zeit} = geneticpls(dat,B,refl,k)
**   Input:      dat (n,p)-matrix, the values of the manifest variables 
**               B   (G,G) lower triangular matrix describing the interrelations of the latent 
**                   variables: 
**                   b_ij=+1 if eta_i and eta_j are positively correlated
**                   b_ij=-1 if eta_i and eta_j are negatively correlated
**                   b_ij= 0 if eta_i does not depend on eta_j in a direct way
**                   (b_ii = 0 !) 
**               refl   (p,1) vector describing with which latent variable the manifest variables
**                            are related.
**                            integers from 1 to G; g at position i means, that the i-th manifest 
**                            variable is connected with the g-th latent variable   
**               k        scalar, the number of clusters to be found  
**   Output:     member   (n,1)-vector, indicator of membership 
**               fitness  (2,1)-vector, values of the global criterion for the clusterwise  
**                        pls models at the beginning and at the end (sum of squared residuals)
**               zeit     time in seconds used to complete the task 
**   Globals:    _pn      size of population   
**               _pC      probability of recombination
**               _pM      probability of mutation
**               _limit   number of generations
**   Remarks:    depends on popfit, plspfad, rempiric          
**
*/


proc (3) = geneticpls(dat,B,refl,k);

local pn,pC,pM,limit; 
local popsize, optfit, fitness, optindv, popneu,cross,mut;
local i,fit,j,x1,y1,r,pred,gen,ind,pop0,selprob,u,abc,n;
local startzeit,endzeit;
local B1,dat1,xi,w,lambda,resa,resi,ret,ret1,popi,p,minzahl; 

startzeit = (3600~60~1~0.01)*time;

pn    = _pn;   
pC    = _pC;   
pM    = _pM;   
limit = _limit;

p = cols(dat); 
abc = seqa(1,1,k)';
minzahl = minc(sumc(abs(refl).==abc));
n = rows(dat);                                   // sample size
ind = seqa(1,1,n); 
 
popneu = rempiric(n,pn,abc',1);                // start population    
  
i = 0;
do while i < pn;                 
   i = i+1;  
   if (sumc(popneu[.,i].==abc) < minzahl); 
      popneu[.,i] = rempiric(n,1,abc',1);  
      i = i-1;
   endif;
endo;                                         
   
fitness = popfit(dat,B,refl,k,popneu[.,1:pn]);   // determinig fitness of members of population
optfit = minc(fitness);                          // optimal fitness of the first generation
optindv = minindc(fitness);                      // individual with optimal fitness  

/* main loop */
 
gen = 0; 
do while gen < limit;                               // forming of the next generation   

   gen = gen + 1;  
   pop0 = popneu;           
   selprob = n+1-rankindx(fitness,1);  
   selprob = 0|cumsumc(selprob)/sumc(selprob);      // probability to get selected is invers pro-
                                                    // portional to the sum of the squared residuals
                                                        
   /*  building the new generation  */
     
   popneu = pop0[.,optindv];                        // the fittest individual is selected
   
   aufbau:                                          // genetic manipulations
   u = rndu(1,1);
   popsize = cols(popneu);
   
   if ((u <= pC) and (popsize <= (pn-2)));          // recombination
   
      r = rndu(4,1);
      i = sumc(r[1 2]' .> selprob );
      cross = pop0[.,i];
      i = sortc( ceil(r[3 4]*n),1 );
      cross[i[1]:i[2],1 2] = cross[i[1]:i[2],2 1];
      if (sumc(cross[.,1].==abc) >= minzahl)*(sumc(cross[.,2].==abc) >= minzahl);
        popneu = popneu~cross;   
      endif;
   elseif ( u <= (pC+pM));                          // mutation
   
      r = rndu(3,1);
      i = sumc(r[1] .> selprob);
      mut = pop0[.,i];
      i = ceil(r[2]*n);
      mut[i] = ceil(r[3]*k); 
      if (sumc(mut.==abc) >= minzahl);
         popneu = popneu~mut; 
      endif;
      
   else;                                            // reproduction       
    
      i = sumc(rndu(1,1) .> selprob);
      popneu = popneu~pop0[.,i]; 
   
   endif;
    
     
   if cols(popneu) < pn;                            // check, if new population is large enough  
     goto aufbau;                                   // enlarging if necessary 
   endif; 
   
                                                    // determinig fitness of members of population 
   fitness = popfit(dat,B,refl,k,popneu[.,1:pn]); 
   optindv = minindc(fitness);                      // individual with optimal fitness  
 
endo;

optfit = optfit|minc(fitness);                      // optimal fitness at the begining and at the end 
endzeit = (3600~60~1~0.01)*time;
    
retp(popneu[.,optindv],optfit,endzeit-startzeit);    
endp;

/*
**> popfit
**
**   Purpose:    help function for geneticpls 
** 
**   Format:     fitness = popfit(dat,B,refl,k,popu)
**   Input:      dat (n,p)-matrix, the values of the manifest variables 
**               B   (G,G) lower triangular matrix describing the interrelations of the latent 
**                   variables: 
**                   b_ij=+1 if eta_i and eta_j are positively correlated
**                   b_ij=-1 if eta_i and eta_j are negatively correlated
**                   b_ij= 0 if eta_i does not depend on eta_j in a direct way
**                   (b_ii = 0 !) 
**               refl   (p,1) vector describing with which latent variable the manifest variables
**                            are related.
**                            integers from 1 to G; g at position i means, that the i-th manifest 
**                            variable is connected with the g-th latent variable   
**               k        scalar, the number of clusters to be found  
**               popu   (n,pn)- matrix with elements from {1,...,k} each column corresponds with one chromosome 
**   Output:     fitness  (pn,1)-vector, values of the global criterion for the clusterwise  
**                        pls models for each chromosome          
**
*/

proc (1) = popfit(dat,B,refl,k,popu);

local i,ret,n,fit,j,dat1,ret1,minzahl,B1,xi,w,lambda,resa,resi,pnteil,fitness,R2 ; 
pnteil = cols(popu); 
fitness = zeros(pnteil,1);
i = 0;
do while i < pnteil;                                    
   i = i+1;
   ret = {};
   fit = 0; 
   for j(1,k,1);
     dat1 = selif(dat,popu[.,i].==j); 
     ret1 = 1; 
       {B1,xi,w,lambda,resa,resi,R2,ret1} = plspfad(dat1,B,refl);  
       fit = fit + sumc(vec(resi^2));               // fitness, the sum of squared residuals  
     ret = ret1;
   endfor;  
   if ret == ret|ret1; 
     fitness[i] = fit;
   else;  
    i = i-1; 
   endif; 
endo;
retp(fitness);
endp;
 
 
/*
**> geneticplsThread :  same as  geneticpls but with 4 threads        
**
*/


proc (3) = geneticplsThread(dat,B,refl,k);

local pn,pC,pM,limit; 
local popsize, optfit, fitness, optindv, popneu,cross,mut;
local i,fit,j,x1,y1,r,pred,gen,ind,pop0,selprob,u,abc,n;
local startzeit,endzeit;
local B1,dat1,xi,w,lambda,resa,resi,ret,ret1,popi,p,minzahl;
local pnteil,fitness1,fitness2,fitness3,fitness4;

startzeit = (3600~60~1~0.01)*time;

pn    = _pn;   
pC    = _pC;   
pM    = _pM;   
limit = _limit;

p = cols(dat); 
abc = seqa(1,1,k)';
minzahl = minc(sumc(abs(refl).==abc));
n = rows(dat);                                   // sample size
ind = seqa(1,1,n); 
 
popneu = rempiric(n,pn,abc',1);                // start population    
  
i = 0;
do while i < pn;                 
   i = i+1;  
   if (sumc(popneu[.,i].==abc) < minzahl); 
      popneu[.,i] = rempiric(n,1,abc',1);  
      i = i-1;
   endif;
endo;     

pnteil = pn/4;                                   // determinig fitness of members of population
   
ThreadStat   fitness1 = popfit(dat,B,refl,k,popneu[.,1:pnteil]);
ThreadStat   fitness2 = popfit(dat,B,refl,k,popneu[.,(pnteil+1):(2*pnteil)]);
ThreadStat   fitness3 = popfit(dat,B,refl,k,popneu[.,(2*pnteil+1):(3*pnteil)]);
ThreadStat   fitness4 = popfit(dat,B,refl,k,popneu[.,(3*pnteil+1):pn]);
ThreadJoin;   

fitness = fitness1|fitness2|fitness3|fitness4; 
optfit = minc(fitness);                          // optimal fitness of the first generation
optindv = minindc(fitness);                      // individual with optimal fitness  

/* main loop */
 
gen = 0; 
do while gen < limit;                               // forming of the next generation   

   gen = gen + 1;  
   pop0 = popneu;           
   selprob = n+1-rankindx(fitness,1);  
   selprob = 0|cumsumc(selprob)/sumc(selprob);      // probability to get selected is invers pro-
                                                    // portional to the sum of the squared residuals
                                                        
   /*  building the new generation  */
     
   popneu = pop0[.,optindv];                        // the fittest individual is selected
   
   aufbau:                                          // genetic manipulations
   u = rndu(1,1);
   popsize = cols(popneu);
   
   if ((u <= pC) and (popsize <= (pn-2)));          // recombination
   
      r = rndu(4,1);
      i = sumc(r[1 2]' .> selprob );
      cross = pop0[.,i];
      i = sortc( ceil(r[3 4]*n),1 );
      cross[i[1]:i[2],1 2] = cross[i[1]:i[2],2 1];
      if (sumc(cross[.,1].==abc) >= minzahl)*(sumc(cross[.,2].==abc) >= minzahl);
        popneu = popneu~cross;   
      endif;
   elseif ( u <= (pC+pM));                          // mutation
   
      r = rndu(3,1);
      i = sumc(r[1] .> selprob);
      mut = pop0[.,i];
      i = ceil(r[2]*n);
      mut[i] = ceil(r[3]*k); 
      if (sumc(mut.==abc) >= minzahl);
         popneu = popneu~mut; 
      endif;
      
   else;                                            // reproduction       
    
      i = sumc(rndu(1,1) .> selprob);
      popneu = popneu~pop0[.,i]; 
   
   endif;
    
     
   if cols(popneu) < pn;                            // check, if new population is large enough  
     goto aufbau;                                   // enlarging if necessary 
   endif; 
   
                                                    // determinig fitness of members of population 
   ThreadStat   fitness1 = popfit(dat,B,refl,k,popneu[.,1:pnteil]);
   ThreadStat   fitness2 = popfit(dat,B,refl,k,popneu[.,(pnteil+1):(2*pnteil)]);
   ThreadStat   fitness3 = popfit(dat,B,refl,k,popneu[.,(2*pnteil+1):(3*pnteil)]);
   ThreadStat   fitness4 = popfit(dat,B,refl,k,popneu[.,(3*pnteil+1):pn]);
   ThreadJoin;  
   
   fitness = fitness1|fitness2|fitness3|fitness4; 
   optindv = minindc(fitness);                      // individual with optimal fitness  
 
endo;

optfit = optfit|minc(fitness);                      // optimal fitness at the begining and at the end 
endzeit = (3600~60~1~0.01)*time;
    
retp(popneu[.,optindv],optfit,endzeit-startzeit);    
endp;
